/**
 * description
 *
 * @author eson
 *2022年7月12日-13:36:24
 */
package com.yuandian.dataflow.statemachine;
 
import java.io.File;
import java.nio.ByteBuffer;
import java.util.concurrent.Executor;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.ThreadPoolExecutor;

import com.alipay.remoting.NamedThreadFactory;
import com.alipay.remoting.exception.CodecException;
import com.alipay.remoting.serialization.SerializerManager;
import com.alipay.sofa.jraft.JRaftUtils;
import com.alipay.sofa.jraft.Node;
import com.alipay.sofa.jraft.RaftGroupService;
import com.alipay.sofa.jraft.Status;
import com.alipay.sofa.jraft.closure.ReadIndexClosure;
import com.alipay.sofa.jraft.conf.Configuration;
import com.alipay.sofa.jraft.entity.PeerId;
import com.alipay.sofa.jraft.entity.Task;
import com.alipay.sofa.jraft.error.RaftError;
import com.alipay.sofa.jraft.error.RemotingException;
import com.alipay.sofa.jraft.option.CliOptions;
import com.alipay.sofa.jraft.option.NodeOptions;
import com.alipay.sofa.jraft.rpc.impl.BoltRaftRpcFactory;
import com.alipay.sofa.jraft.util.BytesUtil;
import com.alipay.sofa.jraft.util.Endpoint;
import com.alipay.sofa.jraft.util.ThreadPoolUtil;
import com.yuandian.dataflow.statemachine.rpc.SMResponse;
import com.yuandian.dataflow.statemachine.rpc.SyncDataRequest;
import com.yuandian.dataflow.statemachine.rpc.SyncDataProcessor;
import com.yuandian.dataflow.statemachine.state.State;
import com.yuandian.dataflow.statemachine.state.WorkerState;

import lombok.Getter;
import lombok.Setter;
import lombok.var;
import lombok.extern.slf4j.Slf4j;

/**
 * description
 *
 * @author eson
 *2022年7月12日-13:36:24
 */
@Slf4j
@var
public class StateServerFactory {

    private static StateServer ss;
    // 必须初始化
    public static void InitStateServer(String peerstr, Configuration conf) throws Exception {
        if(ss != null) {
            throw new Exception("重复初始化 InitStateServer");
        }
        ss = new StateServerFactory.StateServer(peerstr, conf);

       
        ss.readIndexState(true, new SyncDataClosure() {
            @Override
            public void run(Status status) {
                log.info("add peerid {}", getState());
                var wstate = getState().getWorkers().get(ss.node.getNodeId().getPeerId());
                if(wstate == null) {
                    wstate = new WorkerState();
                    getState().getWorkers().put(ss.node.getNodeId().getPeerId(), wstate);
                    log.info("update: {}", getState());
                    ss.applyState(getState(), new SyncDataClosure() {
                        @Override
                        public void run(Status status) {
                            log.info("{} add workers", ss.node.getNodeId().getPeerId());
                        }
                    } );
                }
            }
            
        });
    }

    // 获取状态服务的对象
    public static StateServer getStateServer() {
        return ss;
    }

    @Getter
    @Setter
    public static class StateServer {
    
    private Node node;
    private RaftGroupService cluster;
    private StateMachine fsm;

    private String groupId = "dataflow";
    private Executor readIndexExecutor = createReadIndexExecutor(); 

    public StateServer(String addr, Configuration conf) {
        String[] peers = new String[]{"localhost:4440","localhost:4441","localhost:4442"};
        String[] sprPeers = new String[]{"3440","3441","3442"};

        // var peeridstr = peers[Integer.parseInt(serverId)];
        // var sprPort = sprPeers[Integer.parseInt(args[0])];

        // String groupId = "jraft";

        // Configuration conf = JRaftUtils.getConfiguration("localhost:4440,localhost:4441,localhost:4442");

        PeerId serverId = JRaftUtils.getPeerId(addr);
        int port = serverId.getPort();

        NodeOptions nodeOptions = new NodeOptions();
        nodeOptions.setElectionTimeoutMs(1000);
        nodeOptions.setSnapshotLogIndexMargin(3600);
        nodeOptions.setInitialConf(conf);

        File RaftDataFile = new File(String.format("./raftdata/%d", port) );
        log.info("{}",RaftDataFile.mkdirs());  

        nodeOptions.setLogUri( String.format("./raftdata/%d/logs", port) );
        nodeOptions.setRaftMetaUri(String.format("./raftdata/%d/rafts", port));
        nodeOptions.setSnapshotUri(String.format("./raftdata/%d/snaps", port));
        fsm = new StateMachine(); // 状态实例初始化
        nodeOptions.setFsm(fsm);

        cluster = new RaftGroupService(groupId, serverId, nodeOptions);
        cluster.getRpcServer().registerProcessor(new SyncDataProcessor());
        node = cluster.start();
        
        if(node.isLeader()) {
             
        }
     
    }

    public boolean isLeader() {
        return this.fsm.isLeader();
    }

    public State getState() {
        return this.fsm.getState();
    }

    public void applyState(State state, SyncDataClosure closure) {
        if (!ss.isLeader()) {
            ss.handlerNotLeaderError(closure);
            return;
        }
        
        try {
            closure.setState(state);
            final Task task = new Task();
            task.setData(ByteBuffer.wrap(SerializerManager.getSerializer(SerializerManager.Hessian2).serialize(state)));
            task.setDone(closure);
            StateServerFactory.getStateServer().getNode().apply(task);
        } catch (CodecException e) {
            String errorMsg = "Fail to encode TaskState";
            log.error(errorMsg, e);
            closure.failure(errorMsg, PeerId.emptyPeer());
            closure.run(new Status(RaftError.EINTERNAL, errorMsg));
        }
    }

    public void readIndexState(final boolean readOnlySafe, final SyncDataClosure closure) {

        if(!readOnlySafe){
            closure.setState(getState());
            closure.success(getState());
            closure.run(Status.OK());
            return;
        }

        getNode().readIndex(BytesUtil.EMPTY_BYTES, new ReadIndexClosure() {
            @Override
            public void run(Status status, long index, byte[] reqCtx) {

                if(status.isOk()){
                    log.info("readIndex {}", getState());
                    closure.setState(getState());
                    closure.success(getState());
                    closure.run(Status.OK());
                    return;
                }
                 
                readIndexExecutor.execute(() -> {
                    if(isLeader()){
                        log.info("Fail to get value with 'ReadIndex': {}, try to applying to the state machine.", status);
                        applyState(getState(), closure);
                    }else {
                        handlerNotLeaderError(closure);
                    }
                });
            }
        });
    }

    public SMResponse redirect() {
        final SMResponse response = new SMResponse();
        response.setSuccess(false);
        if (this.node != null) {
            final PeerId leader = this.node.getLeaderId();
            if (leader != null) {
                response.setRedirect(leader);
            }
        }
        return response;
    }

    public void handlerNotLeaderError(final SyncDataClosure closure) {
        closure.failure("Not leader.", redirect().getRedirect());
        closure.run(new Status(RaftError.EPERM, "Not leader"));
    }


    private Executor createReadIndexExecutor() {
        //        final StoreEngineOptions opts = new StoreEngineOptions();
        //        return StoreEngineHelper.createReadIndexExecutor(opts.getReadIndexCoreThreads());
        
                return ThreadPoolUtil.newBuilder() //
                        .poolName("CounterPool") //
                        .enableMetric(true) //
                        .coreThreads(4) //
                        .maximumThreads(4) //
                        .keepAliveSeconds(60L) //
                        .workQueue(new SynchronousQueue<>()) //
                        .threadFactory(new NamedThreadFactory("CounterService", true)) //
                        .rejectedHandler(new ThreadPoolExecutor.AbortPolicy()) //
                        .build();
            }
}

    public static void main(String[] args) throws InterruptedException, RemotingException {
        var rpcClient = new BoltRaftRpcFactory().createRpcClient();
        rpcClient.init(new CliOptions());
        var resp = rpcClient.invokeSync(new Endpoint("localhost",4441), new SyncDataRequest(), 5000);
        log.info("{}", resp);
    }
}
